//! \file trie.cpp
/*
 .
  ,     .


http://algolist.manual.ru/search/fsearch/tr_trie.zip

-            .
  ..          .
-           .
-       ,
           .
-   {0x0000FFFF,0x00FF00FF,0x0F0F0F0F,0x33333333,...}   

Proteus (lawnmower-man@mail.ru)
*/

#include<stdio.h>
#include<time.h>
#include<windows.h>

#define TEST_MAX   3		//!<      
#define TREE_DEPTH 3		//!<  
#define VAL_COUNT 9000		//!<    

//!   
DWORD Rand()
{
	DWORD res=0;
	for(int n=0; n<4; n++) res=(res<<8)^rand();
	return res;
}

//!  
class Trie{
protected:
	//!  
	struct node_st{
		DWORD value;	//!<     ,     
		node_st *next; 
		node_st *child;
		node_st(DWORD v) { child=next=0; value=v;};
	}*root;
	int Distance(DWORD,DWORD);							//!<    
	bool Find(node_st*,DWORD val,int max,int depth=0);
private:
	DWORD keys[TREE_DEPTH];	//!< 
	void Delete(node_st*);	//!<    
public:
	Trie();
	~Trie();
	void Add(DWORD val);			//!<    
	bool Find(DWORD val,int max);	//!<     
};


Trie::Trie()
{
	root=0;
	for(int n=0; n<TREE_DEPTH; n++) keys[n]=Rand();
}

Trie::~Trie()
{
	Delete(root);
}

void Trie::Delete(node_st *node)
{
	while(node) {
		node_st *next=node->next;
		Delete(node->child);
		delete node;
		node=next;
	}
}

//!       .
//!        
int Trie::Distance(DWORD x,DWORD y)
{
	y=x^=y;
	while(x>>=1) y-=x;
	return (int)y;
}

void Trie::Add(DWORD val)
{
	node_st *node;
	node_st **root=&this->root;
	for(int depth=0; depth<=TREE_DEPTH; depth++) {
		DWORD d=depth<TREE_DEPTH?Distance(val,keys[depth]):val;
		while(*root) {
			node=*root;
			if(node->value==d) break;
			root=&node->next;
		}
		if(!*root) *root=node=new node_st(d);
		if(depth==TREE_DEPTH) return;
		root=&node->child;
	}
}


bool Trie::Find(DWORD val,int max)
{
	return Find(root,val,max);
}

bool Trie::Find(node_st *node,DWORD val,int max,int depth)
{
	if(depth==TREE_DEPTH) {
		while(node) {
			if(Distance(node->value,val)<=max) return true;
			node=node->next;
		}
	}else{
		int d=Distance(val,keys[depth]);
		while(node) {
			if((int)node->value>=d-max && (int)node->value<=d+max && Find(node->child,val,max,depth+1)) return true;
			node=node->next;
		}
	}
	return false;
}

//=========================================================================


//!<  
class TrieDebuger:public Trie
{
	DWORD *array;
	//!   
	struct stat_st{
		int nodes_count;		//!<  
		int leaf_count;			//!<  
		int leaf_node_count;	//!<     
		int last_node_count;	//!<     
	};
	void Dump(node_st*,int);
	bool BruteFind(node_st*,DWORD val,int max,int depth=0);	//!<       
	bool ArrayFind(DWORD val,int max);						//!<       
	void CalcStat(stat_st&,node_st*,int depth=0);			//!<  
public:
	TrieDebuger()  {array=new DWORD[VAL_COUNT];}
	~TrieDebuger() {delete[]array;}
	void Init();			//!<   
	void TestFast();		//!<   
	void TestSlow();		//!<   
	void ShowStat();		//!<  
	void Dump();			//!<    
};

void TrieDebuger::Init()
{
	for(int n=0; n<VAL_COUNT; n++) array[n]=Rand(),Add(array[n]);
}

void TrieDebuger::Dump()
{
	printf("[tree]\n");
	if(VAL_COUNT<30) Dump(root,0);
}

void TrieDebuger::Dump(node_st *node,int depth)
{
	while(node) {
		printf("%*s[%x]\n",depth*3,"",node->value);
		Dump(node->child,depth+1);
		node=node->next;
	}
}

void TrieDebuger::TestFast()
{
	printf("\n[fast test]\n");
	clock_t t1=clock();
	for(int n=0; n<VAL_COUNT; n++) if(!Find(array[n],TEST_MAX)) {
		printf("<error>!\n\a");
		break;
	}
	clock_t t2=clock();
	printf("values count: %d\n",VAL_COUNT);
	printf("tree depth  : %d\n",TREE_DEPTH);
	printf("max distance: %d\n",TEST_MAX);
	printf("time        : %d ms\n",t2-t1);
}

void TrieDebuger::TestSlow()
{
	printf("\n[brute test]\n");
	clock_t t1=clock();
	for(int n=0; n<VAL_COUNT; n++) if(!ArrayFind(array[n],TEST_MAX)) {
		printf("<error>!\n\a");
		break;
	}
	clock_t t2=clock();
	printf("values count: %d\n",VAL_COUNT);
	printf("tree depth  : %d\n",TREE_DEPTH);
	printf("max distance: %d\n",TEST_MAX);
	printf("time        : %d ms\n",t2-t1);
}

//   (     )
bool TrieDebuger::BruteFind(node_st *node,DWORD val,int max,int depth)
{
	while(node) {
		if(depth==TREE_DEPTH) {
			if(Distance(node->value,val)<=max) return true;
		}else {
			if(BruteFind(node->child,val,max,depth+1)) return true;
		}
		node=node->next;
	}
	return false;
}

bool TrieDebuger::ArrayFind(DWORD val,int max)
{
	for(int n=0; n<VAL_COUNT; n++) if(Distance(array[n],val)<=max) return true;
	return true;
}

void TrieDebuger::CalcStat(stat_st &stat,node_st *node,int depth)
{
	while(node) {
		if(depth==TREE_DEPTH-1) stat.last_node_count++;
		if(depth==TREE_DEPTH  ) stat.leaf_count++; 
		else stat.nodes_count++;
		CalcStat(stat,node->child,depth+1);
		node=node->next;
	}
}


void TrieDebuger::ShowStat()
{
	stat_st stat;
	memset(&stat,0,sizeof(stat));
	CalcStat(stat,root);
	stat.leaf_node_count=stat.leaf_count/stat.last_node_count;
	printf("\n[stat]\n");
	printf("   nodes  count    : %d\n",stat.nodes_count);
	printf("   leaf   count    : %d\n",stat.leaf_count);
	printf("leaf in node count : %d\n",stat.leaf_node_count);
}

//=========================================================================


void main()
{
	TrieDebuger obj;
	obj.Init();
	obj.Dump();
	obj.TestFast();
	obj.TestSlow();
	obj.ShowStat();
	//getchar();
}


